// Copyright 1994 by Jon Dart.  All Rights Reserved.

#ifndef _SEARCH_H
#define _SEARCH_H

#include "board.h"
#include "emove.h"
#include "pinfo.h"
#include "hash.h"
#include "timectrl.h"
#include "movegen.h"
#include <time.h>
#ifdef _WINDOWS
#include <windows.h>
#endif

class ReversibleMove;

class Search
{
    // Encapsulates alpha-beta search routine.

    public:
	
    enum StateType {Normal,Terminated,Check,Checkmate,
	            Stalemate,Draw,Resigns};

    struct Statistics
    {
       StateType state;
       int value;
       Move best_line[Constants::MaxPly];
       time_t elapsed_time;
       unsigned plies_completed;
       unsigned max_depth;
       unsigned long num_moves;
       unsigned long num_nodes;
       Statistics();
       void clear();
    };
    
    struct flag_struct
    {
       int in_check : 1;
       int mate : 1;
       int capture_search : 1;
       int finished : 1;
       int forced : 1;
       int pv : 1;
       int extend_search : 1;
       int research: 1; // restarting because of beta cutoff at ply 0
    };

    Search();
    
    ~Search();
    
    void terminate_now();
    
    Move find_best_move( 
#ifdef _WINDOWS
	    CWnd* parentWin,
#endif
	    const Board &ABoard, 
	    const Time_Info &ti,
#ifdef _WINDOWS		    
	    const Boolean background,
#endif		    
	    Statistics &stats,
	    Move &prev_move,
        Boolean use_previous_search,
        Boolean verbose = False );
    
    unsigned hints( const Board &ABoard,
	    Move *moves,
	    unsigned max_moves);
    // fill "moves" with up to "max_moves" hint moves, computed using
    // a minimal search depth.	    
	    
    int get_ply() const;
    // return current depth of search.

    unsigned long get_numnodes() const;
    // return # of nodes visited
	    
    time_t get_start_time() const
    {
        return start_time;
    }
    
    Search_Type get_search_type() const
    {
	return timeCtrl.get_search_type();
    }
    
    Search_Limit get_search_limit() const
    {
	return timeCtrl.get_search_limit();
    }
    
    ColorType get_side_to_move() const
    {
	return side_to_move;
    }
    
    Boolean is_background_search() const
    {
	return bkgrnd_search;
    }
    
    void set_time_up( const Boolean b )
    {
	time_up = b;
    }
    
    Boolean was_terminated() const
    {
	return terminated;
    }
    
    unsigned long get_time_limit() const
    { 
    	return time_target + time_added;
    }

private:
	void init(const Board &);

	// make this class uncopyable:
	Search( const Search & );
	Search &operator = ( const Search & );

	int move_search(Board & ABoard, int alpha, int beta,
	    const int ply, flag_struct flags,
	    Boolean &terminate);
        Boolean skip_move(const Board & ABoard, const unsigned ply,
	  const unsigned limit,
	  const ExtendedMove & emove, const Boolean in_check);
        int try_move(Board & ABoard, const ExtendedMove & emove,
	  const int ply, int alpha, int &beta, const int rank,
          flag_struct flags,
	  Boolean &terminate);
        unsigned generate_moves(
	       Board & board, Move_Generator & mg,
	       const unsigned ply,
	       Move * moves,
	       const flag_struct flags);
        int evalu8(const Board & board);
        int evalu8(const Board & board, int alpha, int beta);
	Boolean is_draw( const Board &board, int &rep_count );
        void iterate(const Board &ABoard, const Search_Limit &limits, Statistics &stats,
                     Boolean verbose);

	unsigned long num_moves;
	unsigned plies_done;
	Move ply0moves[Move_Generator::MaxMoves];
	int ply0scores[Move_Generator::MaxMoves];
	int ply0move_count;
	int rank; // moves completed at ply 0
	Boolean ply0moves_generated;
	ExtendedMove *last_move; // array, size MaxPly
	Move *best_moves; // array, size MaxPly.  Best move at
			  // each ply
        int best_scores[Constants::MaxPly];				   
	Move *pv;		 // array, size MaxPly
	ReversibleMove *move_stack[Constants::MaxPly];
	Move *moves_generated[Constants::MaxPly]; // MaxMoves for each ply
	Boolean forced[Constants::MaxPly];
	Hash *hash_table;
	int extend;
	Search::StateType state;
	ColorType side_to_move;
	Boolean terminated;
	Boolean bkgrnd_search;
	unsigned long time_target, time_added;

#ifdef _WINDOWS
	int search_number;
	static HACCEL accel;
	HWND my_hwnd;
	UINT idTimer;
#endif
	time_t start_time;
	Time_Control timeCtrl;
	unsigned long num_nodes;
	int max_depth; // nominal search limit
	int display_depth; // depth to display
	int max_pos_score;
	Boolean time_up, searching;
#if defined(NULL_MOVE) || defined(DEEP_NULL)
	Boolean null_move;
#endif
        unsigned long hash_searches, hash_hits,
        hash_inserts, hash_replaces, hash_inserts_failed,
	check_extensions, capture_extensions, a, b, c, d, e;
        int root_value, root_beta;
};

#endif

